#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "mrp.h"

extern POINT dyx[];

unsigned int getbits(FILE *fp, int n)
{
    static int bitpos = 0;
    static unsigned int bitbuf = 0;
    int x = 0;

    if (n <= 0) return (0);
    while (n > bitpos) {
	n -= bitpos;
	x = (x << bitpos) | bitbuf;
	bitbuf = getc(fp) & 0xff;
	bitpos = 8;
    }
    bitpos -= n;
    x = (x << n) | (bitbuf >> bitpos);
    bitbuf &= ((1 << bitpos) - 1);
    return (x);
}

DECODER *init_decoder(FILE *fp)
{
    DECODER *dec;

    dec = (DECODER *)malloc(sizeof(DECODER));
    if (getbits(fp, 16) != MAGIC_NUMBER) {
	fprintf(stderr, "Not a compressed file!\n");
	exit(1);
    }
    dec->version = getbits(fp, 8);
    dec->width = getbits(fp, 16);
    dec->height = getbits(fp, 16);
    dec->maxval = getbits(fp, 16);
    dec->num_comp = getbits(fp, 6);
    dec->num_class = getbits(fp, 6);
    dec->num_group = getbits(fp, 6);
    dec->prd_order = getbits(fp, 6);
    dec->coef_precision = getbits(fp, 4) + 1;
    dec->predictor = (int **)alloc_2d_array(dec->num_class, dec->prd_order,
					    sizeof(int));
    dec->th = (int **)alloc_2d_array(dec->num_class, dec->num_group - 1,
				     sizeof(int));
    dec->idx_E = (img_t **)alloc_2d_array(dec->height, dec->width,
					  sizeof(img_t));
    dec->class = (uchar **)alloc_2d_array(dec->height, dec->width,
					  sizeof(uchar));
    dec->vlc = init_vlc(dec->num_group, dec->maxval + 1);
    return (dec);
}

int decode_vlc(FILE *fp, VLC *vlc)
{
    int i, k, min, off;
    unsigned int code;

    code = min = off = k = 0;
    for (i = 0; i < vlc->max_len; i++) {
	code = (code << 1) | getbits(fp, 1);
	k = vlc->off[i];
	if (k < 0) {
	    min <<= 1;
	} else {
	    if (code <= vlc->code[vlc->index[k]]) break;
	    min = (vlc->code[vlc->index[k]] + 1) << 1;
	    off = k + 1;
	}
    }
    i = off + code - min;
    return (vlc->index[i]);
}

int decode_golomb(FILE *fp, int m)
{
    int v = 0;
    while (getbits(fp, 1) == 0) {
	v++;
    }
    v = (v << m) | getbits(fp, m);
    return (v);
}

void decode_class(FILE *fp, DECODER *dec)
{
    int i, j, k, x, y;
    int med[3];
    uchar *sort;
    VLC *vlc;

    sort = (uchar *)malloc(dec->num_class * sizeof(uchar));
    vlc = (VLC *)malloc(sizeof(VLC));
    vlc->size = dec->num_class;
    vlc->max_len = 16;
    vlc->len = (int *)malloc(vlc->size * sizeof(int));
    for (i = 0; i < vlc->size; i++) {
	vlc->len[i] = getbits(fp, 4) + 1;
    }
    gen_huffcode(vlc);
    for (i = 0; i < dec->num_class; i++) {
        sort[i] = i;
    }
    for (y = 0; y < dec->height; y += BLOCK_SIZE) {
	for (x = 0; x < dec->width; x += BLOCK_SIZE) {
	    if (y == 0) {
		if (x == 0) {
                    med[0] = med[1] = med[2] = 0;
		} else {
                    med[0] = med[1] = med[2] = dec->class[y][x-1];
		}
	    } else {
		med[0] = (x == 0)? dec->class[y-1][x] : dec->class[y][x-1];
                med[1] = (x + BLOCK_SIZE >= dec->width)?
		    dec->class[y-1][x] : dec->class[y-1][x+BLOCK_SIZE];
                med[2] = dec->class[y-1][x];
		if (med[0] == med[1]) {
                    med[0] = med[2];
                    med[2] = med[1];
		}
	    }
            /* move to front */
	    for (k = 0; k < 3; k++) {
		for (i = 0; i < dec->num_class; i++) {
                    if (sort[i] == med[k]) break;
		}
		for (j = i; j > 0; j--) {
                    sort[j] = sort[j - 1];
		}
                sort[0] = med[k];
	    }
	    /* conversion */
	    i = decode_vlc(fp, vlc);
	    k = sort[i];
	    for (i = 0; i < BLOCK_SIZE; i++) {
		for (j = 0; j < BLOCK_SIZE; j++) {
		    dec->class[y + i][x + j] = k;
		}
	    }
	}
    }
    free_vlc(vlc);
    free(sort);

    return;
}

void decode_predictor(FILE *fp, DECODER *dec)
{
    int k, m, cl, coef;

    for (k = 0; k < dec->prd_order; k++) {
	m = getbits(fp, 4);
	for (cl = 0; cl < dec->num_class; cl++) {
	    coef = decode_golomb(fp, m);
	    if (coef > 0) {
		if (getbits(fp, 1)) {
		    coef = -coef;
		}
	    }
	    dec->predictor[cl][k] = coef;
	}
    }
    return;
}

void decode_threshold(FILE *fp, DECODER *dec)
{
    int cl, gr, m, k;

    m = getbits(fp, 4);
    for (cl = 0; cl < dec->num_class; cl++) {
	k = 0;
	for (gr = 1; gr < dec->num_group; gr++) {
	    if (k <= MAX_UPARA) {
		k += decode_golomb(fp, m);
	    }
	    dec->th[cl][gr - 1] = k;
	}
    }
    return;
}

int calc_prd(IMAGE *img, DECODER *dec, int cl, int y, int x)
{
    int k, prd, prd_order, rx, ry, *coef_p;

    prd_order = dec->prd_order;
    prd = 0;
    coef_p = dec->predictor[cl];
    if (y == 0) {
	if (x == 0) {
	    for (k = 0; k < prd_order; k++) {
		prd += *coef_p++;
	    }
	    prd *= ((img->maxval + 1) >> 1);
	} else {
	    ry = 0;
	    for (k = 0; k < prd_order; k++) {
		rx = x + dyx[k].x;
		if (rx < 0) rx = 0;
		else if (rx >= x) rx = x - 1;
		prd += (*coef_p++) * img->val[ry][rx];
	    }
	}
    } else {
	if (x == 0) {
	    for (k = 0; k < prd_order; k++) {
		ry = y + dyx[k].y;
		if (ry < 0) ry = 0;
		else if (ry >= y) ry = y - 1;
		rx = x + dyx[k].x;
		if (rx < 0) rx = 0;
		prd += (*coef_p++) * img->val[ry][rx];
	    }
	} else {
	    for (k = 0; k < prd_order; k++) {
		ry = y + dyx[k].y;
		if (ry < 0) ry = 0;
		rx = x + dyx[k].x;
		if (rx < 0) rx = 0;
		else if (rx >= img->width) rx = img->width - 1;
		prd += (*coef_p++) * img->val[ry][rx];
	    }
	}
    }
    prd >>= (dec->coef_precision - 1);
    if (prd < 0) prd = 0;
    else if (prd > (img->maxval << 1)) prd = (img->maxval << 1);

    return (prd);
}

IMAGE *decode_image(FILE *fp, DECODER *dec)
{
    int x, y, cl, gr, prd, u, e, E, p;
    int *th_p;
    IMAGE *img;

    img = alloc_image(dec->width, dec->height, dec->maxval);
    for (y = 0; y < dec->height; y++) {
	for (x = 0; x < dec->width; x++) {
	    cl = dec->class[y][x];
	    u = calc_u(dec->idx_E, dec->width, y, x);
            th_p = dec->th[cl];
	    for (gr = 0; gr < dec->num_group - 1; gr++) {
                if (u < *th_p++) break;
	    }
	    prd = calc_prd(img, dec, cl, y, x);
	    p = (prd + 1) >> 1;
	    dec->idx_E[y][x] = E = decode_vlc(fp, dec->vlc[gr]);
	    e = E2e(E, p, prd & 1, dec->maxval);
	    img->val[y][x] = p + e;
	}
    }
    return (img);
}

void write_pgm(IMAGE *img, char *filename)
{
    int i, j;
    FILE *fp;
    fp = fileopen(filename, "wb");
    fprintf(fp, "P5\n%d %d\n%d\n", img->width, img->height, img->maxval);
    for (i = 0; i < img->height; i++) {
	for (j = 0; j < img->width; j++) {
	    putc(img->val[i][j], fp);
	}
    }
    return;
}

int main(int argc, char **argv)
{
    int i;
    IMAGE *img;
    DECODER *dec;
    clock_t st_clk, ed_clk;
    char *infile, *outfile;
    FILE *fp;

    st_clk = clock();
    setbuf(stdout, 0);
    infile = outfile = NULL;
    for (i = 1; i < argc; i++) {
	if (infile == NULL) {
	    infile = argv[i];
	} else {
	    outfile = argv[i];
	}
    }
    if (infile == NULL || outfile == NULL) {
        printf(BANNER"\n", 0.1 * VERSION);
	printf("usage: decmrp infile outfile\n");
	printf("infile:     Input file\n");
	printf("outfile:    Output file\n");
        exit(0);
    }
    fp = fileopen(infile, "rb");
    dec = init_decoder(fp);
    decode_class(fp, dec);
    decode_predictor(fp, dec);
    decode_threshold(fp, dec);
    img = decode_image(fp, dec);
    fclose(fp);
    write_pgm(img, outfile);
    ed_clk = clock();
    printf("cpu time :%.2f sec.\n", (double)(ed_clk-st_clk) / CLOCKS_PER_SEC);
    return(0);
}