#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include "mrp.h"

extern POINT dyx[];
extern double sigma[];

IMAGE *read_pgm(char *filename)
{
    int i, j, width, height, maxval;
    char tmp[256];
    IMAGE *img;
    FILE *fp;

    fp = fileopen(filename, "rb");
    fgets(tmp, 256, fp);
    if (tmp[0] != 'P' || tmp[1] != '5') {
	fprintf(stderr, "Not a PGM file!\n");
	exit(1);
    }
    while (*(fgets(tmp, 256, fp)) == '#');
    sscanf(tmp, "%d %d", &width, &height);
    while (*(fgets(tmp, 256, fp)) == '#');
    sscanf(tmp, "%d", &maxval);
    if ((width % BLOCK_SIZE) || (height % BLOCK_SIZE)) {
	fprintf(stderr, "Image width and height must be multiples of %d!\n",
		BLOCK_SIZE);
	exit(1);
    }
    if (maxval != 255) {
	fprintf(stderr, "Sorry, this version only supports 8bpp images!\n");
	exit(1);

    }
    img = alloc_image(width, height, maxval);
    for (i = 0; i < img->height; i++) {
	for (j = 0; j < img->width; j++) {
	    img->val[i][j] = (unsigned char)fgetc(fp);
	}
    }
    fclose(fp);
    return (img);
}

img_t ***init_ref_pels(IMAGE *img, int prd_order)
{
    img_t ***ref_pels, *ptr;
    int x, y, rx, ry, k;

    ref_pels = (img_t ***)alloc_2d_array(img->height, img->width,
					 sizeof(img_t *));
    ptr = (img_t *)malloc(sizeof(img_t) * img->height * img->width
			  * prd_order);
    for (y = 0; y < img->height; y++) {
	for (x = 0; x < img->width; x++) {
	    ref_pels[y][x] = ptr;
	    ptr += prd_order;
	}
    }
    for (y = 0; y < img->height; y++) {
	for (x = 0; x < img->width; x++) {
	    ptr = ref_pels[y][x];
	    if (y == 0) {
		if (x == 0) {
		    for (k = 0; k < prd_order; k++) {
			ptr[k] = ((img->maxval + 1) >> 1);
		    }
		} else {
		    ry = 0;
		    for (k = 0; k < prd_order; k++) {
			rx = x + dyx[k].x;
			if (rx < 0) rx = 0;
			else if (rx >= x) rx = x - 1;
			ptr[k] = img->val[ry][rx];
		    }
		}
	    } else {
		if (x == 0) {
		    for (k = 0; k < prd_order; k++) {
			ry = y + dyx[k].y;
			if (ry < 0) ry = 0;
			else if (ry >= y) ry = y - 1;
			rx = x + dyx[k].x;
			if (rx < 0) rx = 0;
			ptr[k] = img->val[ry][rx];
		    }
		} else {
		    for (k = 0; k < prd_order; k++) {
			ry = y + dyx[k].y;
			if (ry < 0) ry = 0;
			rx = x + dyx[k].x;
			if (rx < 0) rx = 0;
			else if (rx >= img->width) rx = img->width - 1;
			ptr[k] = img->val[ry][rx];
		    }
		}
	    }
	}
    }
    return (ref_pels);
}

ENCODER *init_encoder(IMAGE *img, int num_class, int num_group,
		      int prd_order, int coef_precision)
{
    ENCODER *enc;
    int x, y, i, j, k;

    enc = (ENCODER *)malloc(sizeof(ENCODER));
    enc->height = img->height;
    enc->width = img->width;
    enc->maxval = img->maxval;
    enc->num_class = num_class;
    enc->num_group = num_group;
    enc->prd_order = prd_order;
    enc->coef_precision = coef_precision;
    enc->predictor = (int **)alloc_2d_array(enc->num_class, enc->prd_order,
					    sizeof(int));
    enc->th = (int **)alloc_2d_array(enc->num_class, enc->num_group - 1,
				     sizeof(int));
    for (i = 0; i < enc->num_class; i++) {
	for (j = 0; j < enc->num_group - 1; j++) {
	    enc->th[i][j] = 0;
	}
    }
    enc->cost = (int **)alloc_2d_array(enc->num_group, enc->maxval + 1,
    				       sizeof(int));
    enc->upara = (int **)alloc_2d_array(enc->height, enc->width, sizeof(int));
    enc->ref_pels = init_ref_pels(img, enc->prd_order);
    enc->abs_e = (img_t **)alloc_2d_array(enc->height, enc->width,
					  sizeof(img_t));
    enc->idx_E = (img_t **)alloc_2d_array(enc->height, enc->width,
					  sizeof(img_t));
    enc->cost_e = NULL;
    enc->econv = (img_t **)alloc_2d_array(enc->maxval+1, (enc->maxval<<1)+1,
					  sizeof(img_t));
    for (i = 0; i <= enc->maxval; i++) {
	for (j = 0; j <= (enc->maxval << 1); j++) {
	    k = (j + 1) >> 1;
	    enc->econv[i][j] = e2E(i - k, k, j&1, enc->maxval);
	}
    }
    enc->class = (uchar **)alloc_2d_array(enc->height, enc->width,
					  sizeof(uchar));
    enc->group = (uchar **)alloc_2d_array(enc->height, enc->width,
					  sizeof(uchar));
    for (y = 0; y < enc->height; y++) {
	for (x = 0; x < enc->width; x++) {
		enc->group[y][x] = 0;
	}
    }
    enc->vlc = init_vlc(enc->num_group, enc->maxval + 1);
    return (enc);
}

void init_class(IMAGE *img, ENCODER *enc)
{
    int k, x, y, i, j, v, cl, sum, num_block;
    int *var, *tmp, **ptr;

    num_block = enc->height * enc->width / (BLOCK_SIZE * BLOCK_SIZE);
    var = (int *)malloc(num_block * sizeof(int));
    ptr = (int **)malloc(num_block * sizeof(int *));
    for (k = 0; k < num_block; k++) {
	y = (k / (enc->width / BLOCK_SIZE)) * BLOCK_SIZE;
	x = (k % (enc->width / BLOCK_SIZE)) * BLOCK_SIZE;
	var[k] = sum = 0;
	for (i = 0; i < BLOCK_SIZE; i++) {
	    for (j = 0; j < BLOCK_SIZE; j++) {
		v = img->val[y + i][x + j];
		sum += v;
		var[k] += v * v;
	    }
	}
	var[k] -= sum * sum / (BLOCK_SIZE * BLOCK_SIZE);
	ptr[k] = &(var[k]);
    }
    /* sort */
    for (i = num_block - 1; i > 0; i--) {
	for (j = 0; j < i; j++) {
	    if (*ptr[j] > *ptr[j + 1]) {
		tmp = ptr[j];
		ptr[j] = ptr[j + 1];
		ptr[j + 1] = tmp;
	    }
	}
    }
    for (k = 0; k < num_block; k++) {
	cl = (k * enc->num_class) / num_block;
	v = (int)(ptr[k] - var);
	y = (v / (enc->width / BLOCK_SIZE)) * BLOCK_SIZE;
	x = (v % (enc->width / BLOCK_SIZE)) * BLOCK_SIZE;
	for (i = 0; i < BLOCK_SIZE; i++) {
	    for (j = 0; j < BLOCK_SIZE; j++) {
		enc->class[y + i][x + j] = cl;
	    }
	}
    }
    free(ptr);
    free(var);
}

void set_cost_model(ENCODER *enc, int f_mmse)
{
    int gr, e;
    double a, b;

    enc->cost_e = enc->abs_e;
    for (gr = 0; gr < enc->num_group; gr++) {
	if (f_mmse) {
	    a = 0;
	    b = 1.0;
	} else {
	    a = 0.5 * log(2 * M_PI * sigma[gr] * sigma[gr]) / log(2.0);
	    b = 1.0 / (2.0 * log(2.0) * sigma[gr] * sigma[gr]);
	}
	for (e = 0; e <= enc->maxval; e++) {
	    enc->cost[gr][e] = a + b * (double)(e * e) + 0.5;
	}
    }
}

void set_cost_rate(ENCODER *enc)
{
    int gr, E, *len;

    enc->cost_e = enc->idx_E;
    for (gr = 0; gr < enc->num_group; gr++) {
	len = enc->vlc[gr]->len;
	for (E = 0; E <= enc->maxval; E++) {
	    enc->cost[gr][E] = len[E];
	}
    }
}

int predict_region(IMAGE *img, ENCODER *enc,
		    int tly, int tlx, int bry, int brx)
{
    int x, y, k, u, cl, gr, cost, prd, org;
    int *coef_p, *th_p, *upara_p;
    img_t *ref_p, **ref_pp, *org_p, *idx_E_p, *abs_e_p, *cost_e_p;
    uchar *class_p, *group_p;

    for (y = tly; y < bry; y++) {
	class_p = &enc->class[y][tlx];
	org_p = &img->val[y][tlx];
	ref_pp = &enc->ref_pels[y][tlx];
	idx_E_p = &enc->idx_E[y][tlx];
	abs_e_p = &enc->abs_e[y][tlx];
	for (x = tlx; x < brx; x++) {
	    cl = *class_p++;
	    org = *org_p++;
	    ref_p = *ref_pp++;
	    coef_p = enc->predictor[cl];
	    prd = 0;
	    for (k = 0; k < enc->prd_order; k++) {
		prd += (*coef_p++) * (*ref_p++);
	    }
	    prd >>= (enc->coef_precision - 1);
	    if (prd < 0) prd = 0;
	    else if (prd > (enc->maxval << 1)) prd = (enc->maxval << 1);
	    *idx_E_p++ = enc->econv[org][prd];
	    prd = (prd + 1) >> 1;
	    *abs_e_p++ = (org > prd)? (org - prd) : (prd - org);
	}
    }

    bry += UPEL_DIST;
    if (bry > enc->height) bry = enc->height;
    tlx -= UPEL_DIST;
    if (tlx < 0) tlx = 0;
    brx += UPEL_DIST;
    if (brx > enc->width) brx = enc->width;
    cost = 0;
    for (y = tly; y < bry; y++) {
	class_p = &enc->class[y][tlx];
	group_p = &enc->group[y][tlx];
	upara_p = &enc->upara[y][tlx];
	cost_e_p = &enc->cost_e[y][tlx];
	for (x = tlx; x < brx; x++) {
	    *upara_p++ = u = calc_u(enc->idx_E, enc->width, y, x);
	    cl = *class_p++;
	    th_p = enc->th[cl];
	    for (gr = 0; gr < enc->num_group - 1; gr++) {
		if (u < *th_p++) break;
	    }
	    *group_p++ = gr;
	    cost += enc->cost[gr][*cost_e_p++];
	}
    }
    return (cost);
}

int design_predictor(IMAGE *img, ENCODER *enc, int f_mmse)
{
    double **mat, *weight, w, e, d, pivot;
    int x, y, i, j, k, cl, gr, pivpos, *index;
    img_t *ref_p;

    mat = (double **)alloc_2d_array(enc->prd_order, enc->prd_order + 1,
				    sizeof(double));
    index = (int *)malloc(sizeof(int) * enc->prd_order);
    weight = (double *)malloc(sizeof(double) * enc->num_group);
    for (gr = 0; gr < enc->num_group; gr++) {
	if (f_mmse) {
	    weight[gr] = 1.0;
	} else {
	    weight[gr] = 1.0 / (sigma[gr] * sigma[gr]);
	}
    }
    for (cl = 0; cl < enc->num_class; cl++) {
	for (i = 0; i < enc->prd_order; i++) {
	    for (j = 0; j <= enc->prd_order; j++) {
		mat[i][j] = 0.0;
	    }
	}
	for (y = 0; y < enc->height; y++) {
	    for (x = 0; x < enc->width; x++) {
		if (enc->class[y][x] == cl) {
		    gr = enc->group[y][x];
		    ref_p = enc->ref_pels[y][x];
		    for (i = 0; i < enc->prd_order; i++) {
			w = weight[gr] * ref_p[i];
			for (j = i; j < enc->prd_order; j++) {
			    mat[i][j] += w * ref_p[j];
			}
			mat[i][enc->prd_order] += w * img->val[y][x];
		    }
		}
	    }
	}
	for (i = 0; i < enc->prd_order; i++) {
	    index[i] = i;
	    for (j = 0; j < i; j++) {
		mat[i][j] = mat[j][i];
	    }
	}
	for (i = 0; i < enc->prd_order; i++) {
	    pivpos = i;
	    pivot = fabs(mat[index[i]][i]);
	    for (k = i + 1; k < enc->prd_order; k++) {
		if (fabs(mat[index[k]][i]) > pivot) {
		    pivot = fabs(mat[index[k]][i]);
		    pivpos = k;
		}
	    }
	    k = index[i];
	    index[i] = index[pivpos];
	    index[pivpos] = k;
	    if (pivot > 1E-10) {
		d = mat[index[i]][i];
		for (j = i; j <= enc->prd_order; j++) {
		    mat[index[i]][j] /= d;
		}
		for (k = 0; k < enc->prd_order; k++) {
		    if (k == i) continue;
		    d = mat[index[k]][i];
		    for (j = i; j <= enc->prd_order; j++) {
			mat[index[k]][j] -= d * mat[index[i]][j];
		    }
		}
	    }
	}
	w = (1 << enc->coef_precision);
#if 0
	for (i = 0; i < enc->prd_order; i++) {
	    if (fabs(mat[index[i]][i]) > 1E-10) {
		d = mat[index[i]][enc->prd_order] * w + 0.5;
	    } else {
		d = 0.0;
	    }
	    k = d;
	    if (k > d) k--;
	    enc->predictor[cl][i] = k;
	}
#else
	e = 0.0;
	for (i = 0; i < enc->prd_order; i++) {
	    if (fabs(mat[index[i]][i]) > 1E-10) {
                d = mat[index[i]][enc->prd_order] * w;
	    } else {
                d = 0.0;
	    }
            k = d;
            if (k > d) k--;
            enc->predictor[cl][i] = k;
	    d -= k;
	    e += d;
	    mat[index[i]][enc->prd_order] = d;
	}
	/* minimize mean rounding errors */
	k = e + 0.5;
	for (;k > 0; k--) {
	    d = 0;
	    for (j = i = 0; i < enc->prd_order; i++) {
		if (mat[index[i]][enc->prd_order] > d) {
		    d = mat[index[i]][enc->prd_order];
		    j = i;
		}
	    }
	    enc->predictor[cl][j]++;
	    mat[index[j]][enc->prd_order] = 0;
	}
#endif
    }
    free(weight);
    free(index);
    free(mat);
    return (predict_region(img, enc, 0, 0, enc->height, enc->width));
}

int optimize_group(ENCODER *enc)
{
    int x, y, th1, th0, k, u, cost, cl, gr;
    int **cbuf, **trellis, *dpcost;

    cbuf = (int **)alloc_2d_array(enc->num_group, MAX_UPARA + 2, sizeof(int));
    trellis = (int **)alloc_2d_array(enc->num_group, MAX_UPARA + 2, 
				     sizeof(int));

    dpcost = trellis[0];
    /* Dynamic programming */
    for (cl = 0; cl < enc->num_class; cl++) {
	for (gr = 0; gr < enc->num_group; gr++) {
	    for (u = 0; u < MAX_UPARA + 2; u++) {
		cbuf[gr][u] = 0;
	    }
	}
	for (y = 0; y < enc->height; y++) {
	    for (x = 0; x < enc->width; x++) {
		if (enc->class[y][x] == cl) {
		    u = enc->upara[y][x] + 1;
		    for (gr = 0; gr < enc->num_group; gr++) {
			cbuf[gr][u] += enc->cost[gr][enc->cost_e[y][x]];
		    }
		}
	    }
	}
	for (gr = 0; gr < enc->num_group; gr++) {
	    for (u = 1; u < MAX_UPARA + 2; u++) {
		cbuf[gr][u] += cbuf[gr][u - 1];
	    }
	}
	for (u = 0; u < MAX_UPARA + 2; u++) {
	    dpcost[u] = cbuf[0][u];
	}
	for (gr = 1; gr < enc->num_group; gr++) {
	    for (th1 = MAX_UPARA + 1; th1 >= 0; th1--) {
		th0 = th1;
		for (k = 0; k < th1; k++) {
		    cost = dpcost[k] + cbuf[gr][th1] - cbuf[gr][k]; 
		    if (cost < dpcost[th1]) {
			dpcost[th1] = cost;
			th0 = k;
		    }
		}
		trellis[gr][th1] = th0;
		if (gr == enc->num_group - 1) break;
	    }
	}
	th1 = MAX_UPARA + 1;
	for (gr = enc->num_group - 1; gr > 0; gr--) {
	    th1 = trellis[gr][th1];
	    enc->th[cl][gr - 1] = th1;
	}
    }
    /* renew groups */
    cost = 0;
    for (y = 0; y < enc->height; y++) {
	for (x = 0; x < enc->width; x++) {
	    cl = enc->class[y][x];
	    u = enc->upara[y][x];
	    for (gr = 0; gr < enc->num_group - 1; gr++) {
		if (u < enc->th[cl][gr]) break;
	    }
	    enc->group[y][x] = gr;
	    cost += enc->cost[gr][enc->cost_e[y][x]];
	}
    }
    free(trellis);
    free(cbuf);
    return (cost);
}

int optimize_class(IMAGE *img, ENCODER *enc)
{
    int x, y, i, j, cl, cost, min_cost, min_cl;

    for (y = 0; y < enc->height; y += BLOCK_SIZE) {
	for (x = 0; x < enc->width; x += BLOCK_SIZE) {
	    min_cost = 1E8;
	    min_cl = 0;
	    for (cl = 0; cl < enc->num_class; cl++) {
		for (i = y; i < y + BLOCK_SIZE; i++) {
		    for (j = x; j < x + BLOCK_SIZE; j++) {
			enc->class[i][j] = cl;
		    }
		}
		cost = predict_region(img, enc, y, x,
				      y + BLOCK_SIZE, x + BLOCK_SIZE);
		if (cost < min_cost) {
		    min_cost = cost;
		    min_cl = cl;
		}
	    }
	    for (i = y; i < y + BLOCK_SIZE; i++) {
		for (j = x; j < x + BLOCK_SIZE; j++) {
		    enc->class[i][j] = min_cl;
		}
	    }
	    predict_region(img, enc, y, x, y + BLOCK_SIZE, x + BLOCK_SIZE);
	}
    }
    return (predict_region(img, enc, 0, 0, enc->height, enc->width));
}

void optimize_coef(IMAGE *img, ENCODER *enc, int cl, int pos1, int pos2)
{
#define SEARCH_RANGE 11
    int i, j, k, x, y, df1, df2, gr, org, *coef, *coef_p, *cost_p;
    int prd, prd_f1, prd_f2, max_prd;
    int cbuf[SEARCH_RANGE][SEARCH_RANGE], *cbuf_p;
    img_t *ref_p, *econv_p;

    coef = enc->predictor[cl];
    coef[pos1] -= (SEARCH_RANGE >> 1);
    coef[pos2] -= (SEARCH_RANGE >> 1);
    for (i = 0; i < SEARCH_RANGE; i++) {
	for (j = 0; j < SEARCH_RANGE; j++) {
	    cbuf[i][j] = 0;
	}
    }
    max_prd = (enc->maxval << 1);

    for (y = 0; y < enc->height; y++) {
	for (x = 0; x < enc->width; x++) {
	    if (enc->class[y][x] != cl) {
		x += BLOCK_SIZE - 1;
		continue;
	    }
	    org = img->val[y][x];
	    gr = enc->group[y][x];
	    ref_p = enc->ref_pels[y][x];
	    econv_p = enc->econv[org];
	    cost_p = enc->cost[gr];
	    df1 = ref_p[pos1];
	    df2 = ref_p[pos2];
	    coef_p = coef;
	    prd_f1 = 0;
	    for (k = 0; k < enc->prd_order; k++) {
		prd_f1 += (*ref_p++) * (*coef_p++);
	    }
	    for (i = 0; i < SEARCH_RANGE; i++) {
		prd_f2 = prd_f1;
		cbuf_p = cbuf[i];
		for (j = 0; j < SEARCH_RANGE; j++) {
		    prd = prd_f2  >> (enc->coef_precision - 1);
		    if (prd < 0) prd = 0;
		    else if (prd > max_prd) prd = max_prd;
		    (*cbuf_p++) += cost_p[econv_p[prd]];
		    prd_f2 += df2;
		}
		prd_f1 += df1;
	    }
	}
    }
    x = y = 0;
    for (i = 0; i < SEARCH_RANGE; i++) {
	for (j = 0; j < SEARCH_RANGE; j++) {
	    if (cbuf[i][j] < cbuf[y][x]) {
		y = i;
		x = j;
	    }
	}
    }
    coef[pos1] += y;
    coef[pos2] += x;
}

int optimize_predictor(IMAGE *img, ENCODER *enc)
{
    int cl, k, pos1, pos2;

    for (cl = 0; cl < enc->num_class; cl++) {
	for (k = 0; k < enc->prd_order; k++) {
retry:
	    pos1 = (int)(((long)rand() * enc->prd_order) / (RAND_MAX + 1));
	    pos2 = (int)(((long)rand() * enc->prd_order) / (RAND_MAX + 1));
	    if (pos1 == pos2) goto retry;
	    optimize_coef(img, enc, cl, pos1, pos2);
	}
    }
    return (predict_region(img, enc, 0, 0, enc->height, enc->width));
}

void putbits(FILE *fp, int n, unsigned int x)
{
    static int bitpos = 8;
    static unsigned int bitbuf = 0;

    if (n <= 0) return;
    while (n >= bitpos) {
        n -= bitpos;
	if (n < 32) {
            bitbuf |= ((x >> n) & (0xff >> (8 - bitpos)));
	}
        putc(bitbuf, fp);
        bitbuf = 0;
        bitpos = 8;
    }
    bitpos -= n;
    bitbuf |= ((x & (0xff >> (8 - n))) << bitpos);
}

int write_header(ENCODER *enc, FILE *fp)
{
    int bits;

    putbits(fp, 16, MAGIC_NUMBER);
    bits = 16;
    putbits(fp, 8, VERSION);
    bits += 8;
    putbits(fp, 16, enc->width);
    putbits(fp, 16, enc->height);
    putbits(fp, 16, enc->maxval);
    bits += 16 * 3;
    putbits(fp, 6, 1);	/* number of components (1 = monochrome) */
    putbits(fp, 6, enc->num_class);
    putbits(fp, 6, enc->num_group);
    putbits(fp, 6, enc->prd_order);
    putbits(fp, 4, enc->coef_precision - 1);
    bits += 6 * 4 + 4;
    return (bits);
}

int encode_class(FILE *fp, ENCODER *enc)
{
    int i, j, k, blk, x, y, num_block;
    int bits;
    int med[3], *hist;
    uchar *index, *sort;
    VLC *vlc;

    num_block = enc->height * enc->width / (BLOCK_SIZE * BLOCK_SIZE);
    index = (uchar *)malloc(num_block * sizeof(uchar));
    sort = (uchar *)malloc(enc->num_class * sizeof(uchar));
    hist = (int *)malloc(enc->num_class * sizeof(int));
    for (i = 0; i < enc->num_class; i++) {
	sort[i] = i;
    }
    blk = 0;
    for (y = 0; y < enc->height; y += BLOCK_SIZE) {
	for (x = 0; x < enc->width; x += BLOCK_SIZE) {
	    if (y == 0) {
		if (x == 0) {
		    med[0] = med[1] = med[2] = 0;
		} else {
		    med[0] = med[1] = med[2] = enc->class[y][x-1];
		}
	    } else {
		med[0] = (x == 0)? enc->class[y-1][x] : enc->class[y][x-1];
		med[1] = (x + BLOCK_SIZE >= enc->width)?
			enc->class[y-1][x] : enc->class[y-1][x+BLOCK_SIZE];
		med[2] = enc->class[y-1][x];
		if (med[0] == med[1]) {
		    med[0] = med[2];
		    med[2] = med[1];
		}
	    }
	    /* move to front */
	    for (k = 0; k < 3; k++) {
		for (i = 0; i < enc->num_class; i++) {
		    if (sort[i] == med[k]) break;
		}
		for (j = i; j > 0; j--) {
		    sort[j] = sort[j - 1];
		}
		sort[0] = med[k];
	    }
	    /* conversion */
	    k = enc->class[y][x];
	    for (i = 0; i < enc->num_class; i++) {
		if (sort[i] == k) break;
	    }
	    index[blk++] = i;
        }
    }
    for (i = 0; i < enc->num_class; i++) {
	hist[i] = 0;
    }
    for (blk = 0; blk < num_block; blk++) {
	hist[index[blk]]++;
    }
    vlc = make_vlc(hist, enc->num_class, 16);
    bits = 0;
    for (i = 0; i < enc->num_class; i++) {
	putbits(fp, 4, vlc->len[i] - 1);
	bits += 4;
    }
    for (blk = 0; blk < num_block; blk++) {
	k = index[blk];
	putbits(fp, vlc->len[k], vlc->code[k]);
	bits += vlc->len[k];
    }
    free_vlc(vlc);
    free(hist);
    free(sort);
    free(index);
    return (bits);
}

int encode_golomb(FILE *fp, int m, int v)
{
    int bits, p;

    bits = p = (v >> m) + 1;
    while (p > 32) {
	putbits(fp, 32, 0);
	p -= 32;
    }
    putbits(fp, p, 1);	/* prefix code */
    putbits(fp, m, v);
    return (bits + m);
}

int encode_predictor(FILE *fp, ENCODER *enc)
{
    int cl, coef, sgn, k, m, b, min_m, min_b, bits;

    bits = 0;
    for (k = 0; k < enc->prd_order; k++) {
	min_b = INT_MAX;
	for (m = min_m = 0; m < 16; m++) {
	    b = 0;
	    for (cl = 0; cl < enc->num_class; cl++) {
		coef = enc->predictor[cl][k];
		if (coef < 0) coef = -coef;
		b += (coef >> m) + m + 1;
		if (coef != 0) b++;	/* sign */
	    }
	    if (b < min_b) {
		min_b = b;
		min_m = m;
	    }
	}
	putbits(fp, 4, min_m);
	bits += 4;
	for (cl = 0; cl < enc->num_class; cl++) {
	    coef = enc->predictor[cl][k];
	    sgn = (coef < 0)? 1 : 0;
	    if (coef < 0) coef = -coef;
	    bits += encode_golomb(fp, min_m, coef);
	    if (coef != 0) {
		putbits(fp, 1, sgn);
		bits++;
	    }
	}
    }
    return (bits);
}

int encode_threshold(FILE *fp, ENCODER *enc)
{
    int cl, gr, i, k, m, min_m, min_b, bits;

    min_b = INT_MAX;
    for (m = min_m = 0; m < 16; m++) {
	bits = 0;
	for (cl = 0; cl < enc->num_class; cl++) {
	    k = 0;
	    for (gr = 1; gr < enc->num_group; gr++) {
		i = enc->th[cl][gr - 1] - k;
		bits += (i >> m) + m + 1;
		k += i;
		if (k > MAX_UPARA) break;
	    }
	}
	if (bits < min_b) {
	    min_b = bits;
	    min_m = m;
	}
    }
    putbits(fp, 4, min_m);
    bits = 4;
    for (cl = 0; cl < enc->num_class; cl++) {
	k = 0;
	for (gr = 1; gr < enc->num_group; gr++) {
	    i = enc->th[cl][gr - 1] - k;
	    bits += encode_golomb(fp, min_m, i);
	    k += i;
	    if (k > MAX_UPARA) break;
	}
    }
    return (bits);
}

int encode_error(FILE *fp, ENCODER *enc)
{
    int x, y, E, bits;
    VLC *vlc;

    bits = 0;
    for (y = 0; y < enc->height; y++) {
	for (x = 0; x < enc->width; x++) {
	    E = enc->idx_E[y][x];
	    vlc = enc->vlc[enc->group[y][x]];
	    putbits(fp, vlc->len[E], vlc->code[E]);
	    bits += vlc->len[E];
	}
    }
    return (bits);
}

int main(int argc, char **argv)
{
    int i, j, k, cl, bits, cost, min_cost, **prd_save;
    double rate;
    IMAGE *img;
    ENCODER *enc;
    clock_t st_clk, ed_clk;
    int f_mmse = 0;
    int f_optpred = 0;
    int num_class = NUM_CLASS;
    int num_group = NUM_GROUP;
    int prd_order = PRD_ORDER;
    int coef_precision = COEF_PRECISION;
    int max_iteration = MAX_ITERATION;
    char *infile, *outfile;
    FILE *fp;

    st_clk = clock();
    setbuf(stdout, 0);
    infile = outfile = NULL;
    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-') {
	    switch (argv[i][1]) {
		case 'M':
		    num_class = atoi(argv[++i]);
		    if (num_class <= 0 || num_class > 63) {
			num_class = NUM_CLASS;
		    }
		    break;
		case 'K':
		    prd_order = atoi(argv[++i]);
		    if (prd_order <= 0 || prd_order > 56) {
			prd_order = PRD_ORDER;
		    }
		    break;
		case 'P':
		    coef_precision = atoi(argv[++i]);
		    if (coef_precision <= 0 || coef_precision > 16) {
			coef_precision = COEF_PRECISION;
		    }
		    break;
		case 'I':
		    max_iteration = atoi(argv[++i]);
		    if (max_iteration <= 0) {
			max_iteration = MAX_ITERATION;
		    }
		    break;
		case 'm':
		    f_mmse = 1;
		    break;
		case 'o':
		    f_optpred = 1;
		    break;
		default:
		    fprintf(stderr, "Unknown option: %s!\n", argv[i]);
		    exit (1);
	    }
	} else {
	    if (infile == NULL) {
		infile = argv[i];
	    } else {
		outfile = argv[i];
	    }
	}
    }
    if (infile == NULL || outfile == NULL) {
	printf(BANNER"\n", 0.1 * VERSION);
	printf("usage: encmrp [options] infile outfile\n");
	printf("options:\n");
	printf("    -M num  Number of predictors [%d]\n", num_class);
	printf("    -K num  Prediction order [%d]\n", prd_order);
	printf("    -P num  Precision of prediction coefficients (fractional bits) [%d]\n", coef_precision);
	printf("    -I num  Maximum number of iterations [%d]\n", max_iteration);
	printf("    -m      Use MMSE predictors\n");
	printf("    -o      Further optimization of predictors (experimental)\n");
	printf("infile:     Input file (must be in a raw PGM format)\n");
	printf("outfile:    Output file\n");
	exit(0);
    }
    img = read_pgm(infile);
    fp = fileopen(outfile, "wb");
    k = img->width * img->height;
    if (num_class < 0) {
	num_class = 7.5E-5 * k + 10.5;
	if (num_class > 63) num_class = 63;
    }
    if (prd_order < 0) {
	prd_order = 6.5E-5 * k + 17;
	for (i = 1; i < 7; i++) {
	    if (prd_order < (i+1) * (i+1)) {
		prd_order = i * (i+1);
		break;
	    }
	}
	if (i >= 7) prd_order = 56;
    }
    printf("%s -> %s (%dx%d)\n", infile, outfile, img->width, img->height);
    printf("M = %d, K = %d, P = %d\n", num_class, prd_order, coef_precision);
    enc = init_encoder(img, num_class, num_group, prd_order, coef_precision);
    set_cost_model(enc, f_mmse);
    init_class(img, enc);
    prd_save = (int **)alloc_2d_array(enc->num_class, enc->prd_order,
				      sizeof(int));
    /* 1st loop */
    min_cost = INT_MAX;
    for (i = j = 0; i < max_iteration; i++) {
	printf("[%2d] cost =", i);
	cost = design_predictor(img, enc, f_mmse);
	printf(" %d ->", cost);
	cost = optimize_group(enc);
	printf(" %d ->", cost);
	cost = optimize_class(img, enc);
	printf(" %d", cost);
	if (cost < min_cost) {
	    printf(" *\n");
	    min_cost = cost;
	    j = i;
	    for (cl = 0; cl < enc->num_class; cl++) {
		for (k= 0; k < enc->prd_order; k++) {
		    prd_save[cl][k] = enc->predictor[cl][k];
		}
	    }
	} else {
	    printf("\n");
	}
	if (i - j >= EXTRA_ITERATION) break;
    }
    for (cl = 0; cl < enc->num_class; cl++) {
	for (k= 0; k < enc->prd_order; k++) {
	    enc->predictor[cl][k] = prd_save[cl][k];
	}
    }
    set_cost_rate(enc);
    cost = predict_region(img, enc, 0, 0, enc->height, enc->width);
    printf("cost = %d\n", cost);

    /* 2nd loop */
    min_cost = INT_MAX;
    for (i = j = 0; i < max_iteration; i++) {
	printf("(%2d) cost =", i);
	if (f_optpred) {
	    cost = optimize_predictor(img, enc);
	    printf(" %d ->", cost);
	}
	cost = optimize_group(enc);
	printf(" %d ->", cost);
	cost = optimize_class(img, enc);
	printf(" %d", cost);
	if (cost < min_cost) {
	    printf(" *\n");
	    min_cost = cost;
	    j = i;
	    if (f_optpred) {
		for (cl = 0; cl < enc->num_class; cl++) {
		    for (k= 0; k < enc->prd_order; k++) {
			prd_save[cl][k] = enc->predictor[cl][k];
		    }
		}
	    }
	} else {
	    printf("\n");
	}
	if (f_optpred) {
	    if (i - j >= EXTRA_ITERATION) break;
	} else {
	    if (i > j) break;
	}
    }
    if (f_optpred) {
	for (cl = 0; cl < enc->num_class; cl++) {
	    for (k= 0; k < enc->prd_order; k++) {
		enc->predictor[cl][k] = prd_save[cl][k];
	    }
	}
	predict_region(img, enc, 0, 0, enc->height, enc->width);
	optimize_group(enc);
	optimize_class(img, enc);
    }

    bits = k = write_header(enc, fp);
    printf("header info.\t:%10d bits\n", k);
    bits += k = encode_class(fp, enc);
    printf("class info.\t:%10d bits\n", k);
    bits += k = encode_predictor(fp, enc);
    printf("predictors\t:%10d bits\n", k);
    bits += k = encode_threshold(fp, enc);
    printf("thresholds\t:%10d bits\n", k);
    bits += k = encode_error(fp, enc);
    printf("pred. errors\t:%10d bits\n", k);
    printf("------------------------------\n");
    printf("total\t\t:%10d bits\n", bits);
    rate = (double)bits / (enc->height * enc->width);
    printf("coding rate\t:%10.5f b/p\n", rate);

    putbits(fp, 7, 0);
    fclose(fp);
    ed_clk = clock();
    printf("cpu time :%.2f sec.\n", (double)(ed_clk-st_clk) / CLOCKS_PER_SEC);
    return (0);
}
